local Entity = class:extend("Entitiy")
local RECORDS = RECORDS
local getUpdateCount = getUpdateCount
local Operators = OPERATORS

function Entity:__init__(kwargs)
    self.active = true
    
    kwargs = kwargs or {}
    self._kwargs = kwargs
    
    self.toSet = {"x","y","life","vx","vy","source","w","h","angle"}
    
    self.angle = kwargs.angle or 0
    self._w,self._h = 0,0
    self.x = kwargs.x or 0
    self.y = kwargs.y or 0
    self.h = kwargs.h or 16
    self.w = kwargs.w or 16
    self.ax = kwargs.ax or 0
    self.ay = kwargs.ay or 0
    self.passive = kwargs.passive
    self.elasticity = kwargs.elasticity or 0
    
    self.vx, self.vy = kwargs.vx or 0,kwargs.vy or 0
    
    self.sight = kwargs.sight or 10
    self.collisionType = kwargs.collisionType or kwargs.colType or "touch"
    
    self.source = kwargs.source or "none.png"
    self.name = kwargs.name or "Entity"
    
    self.map = kwargs.map or {}

    self.speed = {x=0,y=0}
    self.weight = 400
    self.gravity = 0
    self.initGravity = 500/2

    self._speed = 100
    self.bounced = {x=false,y=false}
    self.tmpBurnt = false
    
    self.life = kwargs.life or 100
    
    self.old = {x=self.x,y=self.y}
    
    
    local data = {
        self.uuid,
        self.__class,
        self.x,
        self.y,
        self.w,
        self.h,
        self.angle,
        self.source,
        self.life,
        self.parent and self.parent.uuid
    }
    
end

function Entity:getData()
    local parent = self.parent and self.parent.uuid
    local r = {
        self.uuid,
        self.__class,
        self.x,
        self.y,
        self.w,
        self.h,
        self.angle,
        self.source,
        self.life,
        parent
    }
    return r
end

function Entity:getAttacked(attacker)
    self.life = self.life - attacker.attack
end

function Entity:set(var, op, val)
    self[var] = val or op if 1 then return end
    local value = self[var]
    if type(op) == "string" and Operators[op] then
        value = operate(value,op,val)
    else
        value = op
    end
    self[var] = value
    
    return value
end

function Entity:getDisplacement()
    return self.x-self.old.x, self.y-self.old.y
end

function Entity:ignore(item)
    return item.isMini--false
end

function Entity:setX(x)
    self.world:update(self,x,self.y)
    self:set("x",x)
end

function Entity:setY(y)
    self.world:update(self,self.x,y)
    self:set("y",y)
end

function Entity:setPos(x,y)
    self.world:update(self,x,y)
    self:set("x",x)
    self:set("y",y)
end

function Entity:moveTo(...)
    return self:setPos(...)
end

function Entity:applyImpulseY(imp)
    --self.vy = self.vy + imp
    self:set("vy","+",imp)
end

function Entity:applyImpulseX(imp)
    --self.vx = self.vx + imp
    self:set("vx","+",imp)
    
end

function Entity:applyImpulse(impX,impY)
    self:applyImpulseX(impX)
    self:applyImpulseY(impY)
end

function Entity.checkCollision(item, other)
    if not other.isBackground then
        return item.collisionType
    end
end

function Entity:resolveCollision(col, notDo)
    self:changeVelocityByCollisionNormal(col.normal.x, col.normal.y)
    if col.normal.y == -1 then
        if not self.isMini and not self.isPlayer then
                --causes puff on land
                if not self.onGround then--self.vy<-15 then
                    self:createPuff(3,5)
                    media.sfx.player_landed:play()
                end
                self.gravity = 0
                self.vy = 0
                self.onGround = true
            
        end
    else
        self.onGround = false
    end
    
    if notDo then self:changeVelocityByImpulse(col) end
end

function Entity:move(x,y)
    local actualX, actualY, collisions, len = self.world:move(self,x,y,
    self.checkCollision)
    local col
    for i = 1, len do
        col = collisions[i]
        self:resolveCollision(col)
    end
    
    return actualX, actualY
end

function Entity:checkIfOnGround(ny)
  if ny < 0 then self.onGround = true return true 
  else self.onGround = false return false end
end

function Entity:changeVelocityByImpulse(col)
    
        local other = col.other
        if not other.isTile then
            --push
            local y,x = col.normal.y, col.normal.x
            if y<0 and self.vy>other.vy then
              --  other:applyImpulseY(self.vy)
            elseif y<0 and self.vy<other.vy then
       --         self:applyImpulseY(other.vy)
            end
            
            if y>0 and self.vy<other.vy then
           --     other:applyImpulseY(self.vy)
            elseif y>0 and self.vy>other.vy then
         --       self:applyImpulseY(other.vy)
            end
            
            if x<0 and self.vx>other.vx then
                other:applyImpulseX(self.vx)
            elseif x<0 and self.vx<other.vx then
                self:applyImpulseX(other.vx)
            end
            
            if x>0 and self.vx<other.vx then
                other:applyImpulseX(self.vx)
            elseif x>0 and self.vx>other.vx then
                self:applyImpulseX(other.vx)
            end
        end
end

function Entity:changeVelocityByCollisionNormal(nx, ny, bounciness)
  local bounciness = bounciness or self.elasticity
  if not bounciness or bounciness == 0 then
      return
  end
  local vx, vy = self.vx, self.vy

  if (nx < 0 and vx > 0) or (nx > 0 and vx < 0) then
    vx = -vx * bounciness
    self.bounced.x = true
  end
  
  if (ny < 0 and vy > 0) or (ny > 0 and vy < 0) then
    vy = -vy * bounciness
    self.bounced.y = true
  end

  self.vx, self.vy = vx, vy
end

function Entity:chageVelocityByCollision(...)
    return self:changeVelocityByCollisionNormal(...)
end

function Entity:getCenter()
    return self.x + self.w / 2,
         self.y + self.h / 2
end

function Entity:updateGravity(dt)
    
    if not self.gravity or self.gravity==0 then
        self.gravity = self.initGravity
    end
    self:set("gravity",self.gravity+self.weight*dt)
    local gravity = self.gravity
    if self.isCrawling then gravity = 0 end
    self.vy = self.vy + gravity*dt
    if self.onGround and self.vy>0 then
    --    self.vy = 0--??
    end
    if self.onGround then
        self.gravity = self.initGravity
    end
    
end

function Entity:updateVelocity()
    local trash = self.updateGravity and self:updateGravity(dt)
    trash = self.changeVelocityByKeys and self:changeVelocityByKeys(dt)
end

function Entity:update(dt, c)
    c = c or getUpdateCount()
    self.updateCount = self.updateCount + 1
    
    if c~=self.updateCount then
        return
    end
    
    self.tmpBurnt = false
    self.old.x, self.old.y = self.x, self.y
    self.speed = {x=0,y=0}
    self._w, self._h = resizeImage(game:getSource(self.source),self.w,self.h)
    
    self:updateGravity(dt)
    
    local vx, vy = self.x+self.vx*dt,self.y+self.vy*dt
    
    if 1 then --???
        local mx, my = self:move(vx,vy,dt)
        self.x, self.y = mx,my
    end
    --self:getTile()
  --  self:setAll()
end

function Entity:setAll() error()--if m5 then return end
    local var, val
    for i = 1, #self.toSet do
        var = self.toSet[i]
        val = self[var]
        if self["_old"..var] ~= val then
            self["_old"..var] = val
            self:set(var,val)
        end
    end
end

function Entity:drawLife()
    return self.lifeBar:draw()
end

function Entity:draw()

    love.graphics.draw(game:getSource(self.source),self.x,self.y,self.angle,
    self._w,self._h,self.w/2, self.h/2)
    self:drawLife()

end

function Entity:dir(n)
    if n > 0 then return 1 elseif n < 0 then return -1
    else return 0 end
end

function Entity:getDir(...)
    return self:dir(...)
end

function Entity:destroy(color)
    if self.map.client and not continue then--???
        warn("TRIED TO RETURN")--return
    end
    
    --log("destroying "..self.uuid)
    if self.rect then
        self.world:remove(self)
        self.rect = nil
    end
  --  self.source = "none.png"
    self.active = false
    local trash = self.tile and self.tile:unUnit(self)
    if self.map.kill then self.map:kill(self) end
    if not self.passive then
        if not self.isBullet and not self.isPuff and not self.isMini and not self.isDebris then
            if not self.map.isOld then
                if color ~= false then
                   self:spawnDebris(color)
                end
            end
        end
    end
    
    --[[if self.map.host then
        self.map.host:sendToAll("destroyEntity",self.uuid)
    end]]--????
end

function Entity:spawnDebris(color)
    if not self.map.active then
        return
    end
    
    color = color or self.color or "brown"
    
    local area = self.w * self.h
    local debrisNumber = math.floor(math.max(15, area / 100))
    

    for i=1, debrisNumber do
        local color = color
        if type(color) == "table" and type(color[1])~=number then
            color = lume.randomchoice(color)
        end
        
        Debris:new({
            map = self.map,
            x = math.random(self.x, self.x + self.w),
            y = math.random(self.y, self.y + self.h),
            color = color
        }
      )
    end
end

function Entity:accelerate(x,y)
    self.vx = self.vx + x
    self.vy = self.vy + y
end  function movee() self.player.world:update(self.player,self.player.x-30,self.player.y) end

function Entity:getTile()
    if self.tile then
        self.tile.color = nil
        self.tile.unit = nil
    end
    
    if not self.active then
        return
    end
    
    local x, y = self.map:convertPixelToTile(self.x,self.y)
    local t = self.map:getTile(x,y)
    if not t then
        
        if self.isBullet then
            self:passiveDestroy()
        end
        return
    end

    local map,pr,t2,t3,t4,t5,tUp = self.map,{}
    t2 = map:getTile(x,y+1) or pr
    t3 = map:getTile(x,y-1) or pr
    t4 = map:getTile(x-1,y) or pr
    t5 = map:getTile(x+1,y) or pr
    tUp = not (t2.isBackground or t3.isBackground or t4.isBackground or t5.isBackground)
    --sometimes when player jumps up to a tile, tiles underneath
    --get dug up
  --  if self.isPlayer then
      --  t.color = getColor("blue") end
    if not self.isDebris and not t.isBackground and tUp and not self.isMini then --.map:getTile(x,y+1).isBackground and not self.isMini then
        local tt
        local des = function(x,y)
            tt=self.map:getTile(x,y)
            if tt and not tt.isBackground then
                tt:destroy()
            end
        end
        for xx = -1,1 do
            if xx == 0 then
                t:destroy()
            else
                des(x+xx,y)
                des(x,y+xx)
                des(x+xx,y+xx)
            end
        end
    end
    
    assert(t)
 --   t:unit(self)
     self.tile = t
     self.tile.unit = self
    
    return self.tile
end

function Entity:createPuff(min, max)
    local min = min or math.random(1,3)
    if max then
        min = math.random(min, max)
    end
    for i = 1,min do
        Puff:new({map=self.map,
            x=self.x + self.w,
            y=self.y + self.h / 2,
            minSize=16 * (1 - math.random()),
            maxSize=39,
            vx=2, vy=3}
        )
        
    end
end

return Entity