Turret = Entity:extend("Turret")

local huge, abs = math.huge, math.abs
local getUpdateCount = getUpdateCount


function Turret:__init__(kwargs)
    self.name = kwargs.name
    self.map = kwargs.map or {world={add=null}}
    self.world = kwargs.world or self.map.world or {}
    kwargs.map = self.map
    kwargs.world = self.world
    
    Entity.__init__(self,kwargs)
    self.source = "turretBox.png"
    self.muzzle = "muzzle.png"
    
    --assert(self.angle>0)
    self.angleY = math.sin(self.angle)
    self.angleX = math.cos(self.angle)
    
    self.delay = kwargs.delay or 0.1

    
    self.ac = kwargs.speed or kwargs.ac or 500
    self.initGravity = kwargs.gravity or self.initGravity
    
    self.hits = kwargs.hits or 1
    --how many seconds bullet lasts
    self.life = kwargs.life or 100
    self.attack = kwargs.attack or 1.5
    
    self.range = kwargs.range or kwargs.sight or 10000
    self.brutal = kwargs.brutal or false
    
    --self.vx = self.ac*self.angleX
    --self.vy = self.ac*-1*self.angleY
    
    --self.angleY, self.angleX=1,1
    self.ignoresParent = true
    self.bullet = {}--bullets.Bullet
    self.bullets = {}
    self.bulletLen = 0
    
end

function Turret:reloadAngles(ang)
    local angle = ang or self.angle
    self.angleY = math.sin(angle)
    self.angleX = math.cos(angle)
end
   

function Turret:update(dt, c)
    if updateG then
        Entity.update(self, dt)
        self:updateGravity(dt)
    end
    
    local target = self:getTarget()
    if target then
        --self:shootAt(target)
        self.angle = self:getAngle(target)
        self:reloadAngles()
    end
    
    if self.life<=0 then
        self:destroy()
    end
end

function Turret:getTarget()
    return self._target--self:getNearest(game.players)
end

local function filter(item)
    if item.isBackground then return false
    end
    return true
end

function Turret:getNearest(objects)
    local distance, nearest, object, tx, ty
    local cx, cy = self:getCenter()
    local shortestDistance = self.range---huge
    for i = 1, #objects do
        object = objects[i]
        distance = self:getSquaredDistance(object)
        
        if distance < shortestDistance then
            tx, ty = object:getCenter()
            local itemInfo, len = self.world:querySegment(cx,cy,tx,ty,filter)
            -- ignore itemsInfo[1] because that's always self
            local info = itemInfo[2]
            if info then assert(not info.isBackground) end
            if (info and info==object) or self.brutal then
                nearest = object
                shortestDistance = distance
                --warn("Target")
            end
        end
    end
    return nearest
end

function Turret:getSquaredDistance(object)
    local dx, dy = self.x - object.x, self.y - object.y
    return dx*dx + dy*dy
end

function Turret:shootAt(target)
    local b = self.bullet:new({
        x = self.x,
        y = self.y,
        angle = self:getAngle(target),
        parent = self,
        map = self.map,
        target = target
    }) --warn()
    table.insert(self.bullets,b)
    self.bulletLen = #self.bullets
end

function Turret:getAngle(object)
    return math.atan2(object.y-self.y, object.x-self.x)*-1
end

function Turret:draw()


    
    local v = game:getSource(self.source)
    local m = game:getSource(self.muzzle)
    local mw, mh = resizeImage(m, self.w, self.h)
    local vw, vh = resizeImage(v, self.w, self.h)
    
    local xx,yy =v:getWidth()/2, v:getHeight()/2
    local mx,my =m:getWidth(), m:getHeight()
    local ca = math.pi^2
    local ang = math.deg(self.angle)
    
    love.graphics.draw(v,self.x,self.y,0,vw,vh)
   love.graphics.draw(m,self.x-self.w/2,self.y+(self.h/2 +(xx*mh)/2),math.pi-(self.angle+ca),
    mw,mh,xx,yy)
    --love.graphics.draw(m,self.x,self.y,self.angle+math.pi*2,mw,mh)--,mx/2,my/2)

end

return Turret

